import { isIntrospectionQuery, logGraphQLOperation } from '@makerx/graphql-core';
import { OperationTypeNode } from 'graphql';
import { omitNil } from '../utils.mjs';

/**
 * This plugin logs GraphQL operations and context creation failure (if specified via options).
 * See options for more details.
 */
function graphqlOperationLoggingPlugin({ logLevel = 'info', contextCreationDidFail, contextCreationFailureLogger, shouldIgnore, ignoreIntrospectionQueries = true, includeResponseData, includeMutationResponseData, adjustVariables, adjustResultData, augmentLogEntry, } = {}) {
    return {
        contextCreationDidFail: async ({ error }) => {
            contextCreationFailureLogger?.error('Context creation failed', { error });
            await contextCreationDidFail?.({ error });
        },
        requestDidStart: ({ contextValue }) => {
            function log(ctx, subsequentPayload) {
                const { started, logger } = contextValue;
                const { operationName, query, variables } = ctx.request;
                const isIntrospection = query && isIntrospectionQuery(query);
                if (isIntrospection && ignoreIntrospectionQueries)
                    return;
                const type = ctx.operation?.operation;
                const result = ctx.response.body.kind === 'single' ? ctx.response.body.singleResult : ctx.response.body.initialResult;
                const errors = result.errors;
                const adjustedVariables = adjustVariables && variables ? adjustVariables(variables) : variables;
                const data = subsequentPayload ? subsequentPayload.incremental : result.data;
                let adjustedData = includeResponseData
                    ? data
                    : includeMutationResponseData && type === OperationTypeNode.MUTATION
                        ? data
                        : undefined;
                if (adjustResultData && adjustedData)
                    adjustedData = adjustResultData(adjustedData);
                const adjustedResult = omitNil({ errors, data: adjustedData });
                const additionalLogEntryProperties = augmentLogEntry ? omitNil(augmentLogEntry(contextValue)) : undefined;
                logGraphQLOperation({
                    logger,
                    logLevel,
                    type,
                    operationName,
                    query,
                    started,
                    variables: adjustedVariables && Object.keys(adjustedVariables).length > 0 ? adjustedVariables : undefined,
                    result: Object.keys(adjustedResult).length ? adjustedResult : undefined,
                    isIntrospectionQuery: isIntrospection || undefined,
                    isIncrementalResponse: ctx.response.body.kind === 'incremental' || undefined,
                    isSubsequentPayload: !!subsequentPayload || undefined,
                    ...additionalLogEntryProperties,
                });
            }
            const responseListener = {
                willSendResponse(ctx) {
                    if (shouldIgnore?.(ctx))
                        return Promise.resolve();
                    log(ctx);
                    return Promise.resolve();
                },
                willSendSubsequentPayload(ctx, payload) {
                    if (shouldIgnore?.(ctx))
                        return Promise.resolve();
                    log(ctx, payload);
                    return Promise.resolve();
                },
            };
            return Promise.resolve(responseListener);
        },
    };
}

export { graphqlOperationLoggingPlugin };
//# sourceMappingURL=graphql-operation-logging-plugin.mjs.map
